#include "GamepadHandler.h"
#include "Logger.h"
#include "MotorControl.h"
#include "Gyro.h"
#include "PIDController.h"

#define STICK_DRIFT_THRESHOLD 50
#define maxRollCorrection 30
#define maxPitchCorrection 30

// Interne Variablen
static ControllerPtr myControllers[BP32_MAX_GAMEPADS] = {nullptr};
static bool holdMode = false;
static int heldMotorSpeed = 0;
static int lastAxisX = 0;
static int lastAxisY = 0;
static unsigned long lastUpdateTime = 0;
static const unsigned long failsafeTimeout = 20000;

// PID-Regler für Pitch und Roll
static PIDController pidPitch(1.0, 0.1, 0.05);
static PIDController pidRoll(1.0, 0.1, 0.05);

static double mapDouble(double x, double in_min, double in_max, double out_min, double out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

static int clampMotorSpeed(int speed) {
  if (speed < 0) return 0;
  if (speed > 255) return 255;
  return speed;
}

// Bluepad32 Callback-Funktionen
void onConnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == nullptr) {
      myControllers[i] = ctl;
      Logger::log("🎮 Controller connected!");
      break;
    }
  }
}

void onDisconnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == ctl) {
      myControllers[i] = nullptr;
      Logger::log("⚠️ Controller disconnected!");
      MotorControl::stopMotors();
      break;
    }
  }
}

void GamepadHandler::init() {
  BP32.setup(&onConnectedController, &onDisconnectedController);
  BP32.forgetBluetoothKeys();
  Logger::log("Bluepad32 initialized");
}

void GamepadHandler::process() {
  if (!BP32.update()) return;
  for (auto ctl : myControllers) {
    if (ctl && ctl->isConnected()) {
      lastUpdateTime = millis();
      
      int axisX = ctl->axisX();
      int axisY = ctl->axisY();
      Logger::log("🎮 Stick: X=" + String(axisX) + ", Y=" + String(axisY));
      
      int throttle = map(ctl->throttle(), 0, 1023, 0, 255);
      
      bool yPressed = (ctl->buttons() & 0x0008);
      int rtPressed = ctl->buttons() & 0x0080;
      if (yPressed && !holdMode) {
        holdMode = true;
        heldMotorSpeed = throttle;
        lastAxisX = axisX;
        lastAxisY = axisY;
        Logger::log("🔒 Hold mode activated! Speed locked at " + String(heldMotorSpeed));
      }
      if (holdMode) {
        throttle = heldMotorSpeed;
      }
      if (holdMode && !yPressed) {
        int deltaX = abs(axisX - lastAxisX);
        int deltaY = abs(axisY - lastAxisY);
        if ((deltaX > STICK_DRIFT_THRESHOLD || deltaY > STICK_DRIFT_THRESHOLD) || (ctl->buttons() & ~0x0008)) {
          holdMode = false;
          heldMotorSpeed = 0;
          Logger::log("🔓 Hold mode deactivated!");
        }
      }
      if (!holdMode && rtPressed > 0) {
        heldMotorSpeed = throttle;
      }
      
      // Additive Lenkungskorrekturen
      int leftCorrection = (axisX < 0) ? map(abs(axisX), 0, 511, 0, maxRollCorrection) : 0;
      int rightCorrection = (axisX > 0) ? map(axisX, 0, 512, 0, maxRollCorrection) : 0;
      int frontCorrection = (axisY < 0) ? map(abs(axisY), 0, 511, 0, maxPitchCorrection) : 0;
      int rearCorrection = (axisY > 0) ? map(axisY, 0, 512, 0, maxPitchCorrection) : 0;
      
      int motorSpeed15 = clampMotorSpeed(throttle - leftCorrection - frontCorrection);
      int motorSpeed22 = clampMotorSpeed(throttle - leftCorrection - rearCorrection);
      int motorSpeed13 = clampMotorSpeed(throttle - rightCorrection - frontCorrection);
      int motorSpeed32 = clampMotorSpeed(throttle - rightCorrection - rearCorrection);
      
      Logger::log("Steering corrections: Left=" + String(leftCorrection) +
                  ", Right=" + String(rightCorrection) +
                  ", Front=" + String(frontCorrection) +
                  ", Rear=" + String(rearCorrection));
      
      // PID-Stabilisierung (bei aktivem Throttle)
      if (throttle > 0) {
        double correctedPitch, correctedRoll;
        Gyro::getCorrectedAngles(correctedPitch, correctedRoll);
        Logger::log("Gyro: Corrected Pitch=" + String(correctedPitch, 2) +
                    ", Roll=" + String(correctedRoll, 2));
        
        unsigned long now = millis();
        static unsigned long lastPIDTime = now;
        double dt = (now - lastPIDTime) / 1000.0;
        lastPIDTime = now;
        
        double pidOutPitch = pidPitch.compute(0, correctedPitch, dt);
        double pidOutRoll  = pidRoll.compute(0, correctedRoll, dt);
        
        int corr15 = (int)(pidOutPitch + pidOutRoll);
        int corr22 = (int)(-pidOutPitch + pidOutRoll);
        int corr13 = (int)(pidOutPitch - pidOutRoll);
        int corr32 = (int)(-pidOutPitch - pidOutRoll);
        
        motorSpeed15 = clampMotorSpeed(motorSpeed15 + corr15);
        motorSpeed22 = clampMotorSpeed(motorSpeed22 + corr22);
        motorSpeed13 = clampMotorSpeed(motorSpeed13 + corr13);
        motorSpeed32 = clampMotorSpeed(motorSpeed32 + corr32);
        
        Logger::log("PID corrections: Pitch=" + String(pidOutPitch, 2) +
                    ", Roll=" + String(pidOutRoll, 2));
      }
      
      MotorControl::updateMotorSpeeds(motorSpeed15, motorSpeed22, motorSpeed13, motorSpeed32);
    }
  }
}

unsigned long GamepadHandler::getLastUpdateTime() {
  return lastUpdateTime;
}
