#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <Bluepad32.h>

// WLAN-Zugangsdaten
const char* ssid = "DeineSSID";   // ⚠️ Ersetze mit deiner WLAN-SSID
const char* password = "DeinPasswort"; // ⚠️ Ersetze mit deinem WLAN-Passwort

// Webserver & WebSocket
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

// PIN CONNECTIONS
#define MOTOR_PIN1 15 // Motor 1: vorne links (vl)
#define MOTOR_PIN2 22 // Motor 2: unten links (ul)
#define MOTOR_PIN3 13 // Motor 3: vorne rechts (vr)
#define MOTOR_PIN4 32 // Motor 4: unten rechts (ur)

// PWM-Einstellungen
int pwmChannel1 = 0, pwmChannel2 = 1, pwmChannel3 = 2, pwmChannel4 = 3;
int pwmFrequency = 5000, pwmResolution = 8;

unsigned long lastUpdateTime = 0;
const unsigned long failsafeTimeout = 20000;
unsigned long lastHeartbeat = 0;

// Controller & Hold-Modus
ControllerPtr myControllers[BP32_MAX_GAMEPADS];
bool holdMode = false;
int heldMotorSpeed = 0;

const int STICK_DRIFT_THRESHOLD = 50;
int lastAxisX = 0, lastAxisY = 0;

// **Hilfsfunktion: Map für Double-Werte**
double mapDouble(double x, double in_min, double in_max, double out_min, double out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

// **Log-Funktion (WebSocket & Serial)**
void logMessage(const String &msg) {
  Serial.println(msg);
  ws.textAll(msg);
}

// **Motoren stoppen**
void stopMotors() {
  ledcWrite(pwmChannel1, 0);
  ledcWrite(pwmChannel2, 0);
  ledcWrite(pwmChannel3, 0);
  ledcWrite(pwmChannel4, 0);
  logMessage("🚨 Motors stopped!");
}

// **Controller-Callback: Verbunden**
void onConnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == nullptr) {
      myControllers[i] = ctl;
      logMessage("🎮 Controller connected!");
      break;
    }
  }
}

// **Controller-Callback: Getrennt**
void onDisconnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == ctl) {
      myControllers[i] = nullptr;
      logMessage("⚠️ Controller disconnected!");
      stopMotors();
      break;
    }
  }
}

// **Controller-Verarbeitung**
void processGamepad(ControllerPtr ctl) {
  lastUpdateTime = millis();

  int axisX = ctl->axisX();  
  int axisY = ctl->axisY();  
  logMessage("🎮 Analog Stick: X=" + String(axisX) + ", Y=" + String(axisY));

  bool yPressed = (ctl->buttons() & 0x0008);

  int throttleVal = ctl->throttle();
  int motorSpeed = map(throttleVal, 0, 1023, 0, 255);

  // **Hold-Modus aktivieren**
  if (yPressed && !holdMode) {
    holdMode = true;
    heldMotorSpeed = motorSpeed;
    lastAxisX = axisX;
    lastAxisY = axisY;
    logMessage("🔒 Hold mode activated! Speed locked at " + String(heldMotorSpeed));
  }

  // **Hold-Modus hält die Geschwindigkeit konstant**
  if (holdMode) {
    motorSpeed = heldMotorSpeed;
  }

  int rtPressed = ctl->buttons() & 0x0080;
  if (!holdMode && rtPressed > 0) {
    heldMotorSpeed = motorSpeed;
  }

  // **Motorsteuerung basierend auf Stick & Geschwindigkeit**
  double leftSpeed  = (axisX <  0) ? mapDouble(axisX, -511.0, 0.0, 0.7, 1.0) : 1.0;
  double rightSpeed = (axisX >= 0) ? mapDouble(axisX, 0.0, 512.0, 1.0, 0.7) : 1.0;
  double upSpeed    = (axisY <  0) ? mapDouble(axisY, -511.0, 0.0, 1.0, 0.7) : 1.0;
  double downSpeed  = (axisY >= 0) ? mapDouble(axisY, 0.0, 512.0, 1.0, 0.7) : 1.0;

  int motorSpeed1 = motorSpeed * leftSpeed * upSpeed;
  int motorSpeed2 = motorSpeed * leftSpeed * downSpeed;
  int motorSpeed3 = motorSpeed * rightSpeed * upSpeed;
  int motorSpeed4 = motorSpeed * rightSpeed * downSpeed;

  motorSpeed1 = constrain(motorSpeed1, 0, 255);
  motorSpeed2 = constrain(motorSpeed2, 0, 255);
  motorSpeed3 = constrain(motorSpeed3, 0, 255);
  motorSpeed4 = constrain(motorSpeed4, 0, 255);

  ledcWrite(pwmChannel1, motorSpeed1);
  ledcWrite(pwmChannel2, motorSpeed2);
  ledcWrite(pwmChannel3, motorSpeed3);
  ledcWrite(pwmChannel4, motorSpeed4);

  logMessage("⚡ Motor Speeds: " + String(motorSpeed1) + ", " + String(motorSpeed2) + ", " + String(motorSpeed3) + ", " + String(motorSpeed4));
}

// **WebSocket-Event**
void onWsEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  if (type == WS_EVT_CONNECT) {
    Serial.printf("✅ WebSocket client #%u connected\n", client->id());
  } else if (type == WS_EVT_DISCONNECT) {
    Serial.printf("❌ WebSocket client #%u disconnected\n", client->id());
  }
}

void setup() {
  Serial.begin(921600);

  WiFi.begin(ssid, password);
  Serial.print("🌐 Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\n✅ Connected! IP: " + WiFi.localIP().toString());
  logMessage("🌐 ESP32 connected! IP: " + WiFi.localIP().toString());

  // **Webserver & WebSocket**
  ws.onEvent(onWsEvent);
  server.addHandler(&ws);
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
    request->send_P(200, "text/html", R"rawliteral(
      <!DOCTYPE html>
      <html>
      <head><meta charset="utf-8"><title>ESP32 Log</title></head>
      <body>
        <h1>ESP32 Log</h1>
        <pre id="log"></pre>
        <script>
          var ws = new WebSocket('ws://' + location.host + '/ws');
          ws.onmessage = function(event) {
            var logElem = document.getElementById('log');
            logElem.innerHTML = event.data + "\n" + logElem.innerHTML;
          };
          setInterval(() => ws.send("ping"), 5000);
        </script>
      </body>
      </html>
    )rawliteral");
  });
  server.begin();

  BP32.setup(&onConnectedController, &onDisconnectedController);
  BP32.forgetBluetoothKeys();
}

void loop() {
  if (BP32.update()) {
    for (auto ctl : myControllers) {
      if (ctl && ctl->isConnected()) {
        processGamepad(ctl);
      }
    }
  }

  if (millis() - lastUpdateTime > failsafeTimeout) {
    logMessage("⏳ Failsafe aktiviert, Motoren gestoppt!");
    stopMotors();
  }

  delay(10);
}
