#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <Bluepad32.h>
#include <Adafruit_MPU6050.h>
#include <Adafruit_Sensor.h>
#include <Wire.h>
#include <math.h>

// ***** Access Point Einstellungen *****
const char* apSSID = "ESP32_Drone";
const char* apPassword = "drone1234";

// ***** Webserver & WebSocket *****
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

// ***** PIN-Definitionen *****
#define MOTOR_PIN1 15  // Motor 15: vorne links
#define MOTOR_PIN2 22  // Motor 22: hinten links
#define MOTOR_PIN3 13  // Motor 13: vorne rechts
#define MOTOR_PIN4 32  // Motor 32: hinten rechts

// ***** PWM-Einstellungen *****
int pwmChannel1 = 0;
int pwmChannel2 = 1;
int pwmChannel3 = 2;
int pwmChannel4 = 3;
int pwmFrequency = 5000;
int pwmResolution = 8;

// ***** Failsafe *****
unsigned long lastUpdateTime = 0; 
const unsigned long failsafeTimeout = 20000;

// ***** Controller & Hold-Modus *****
ControllerPtr myControllers[BP32_MAX_GAMEPADS];
bool holdMode = false;
int heldMotorSpeed = 0;
const int STICK_DRIFT_THRESHOLD = 50;
int lastAxisX = 0;
int lastAxisY = 0;

// ***** PID-Regler (für Pitch & Roll) *****
double Kp = 1.0;    // Proportionalanteil – ggf. anpassen!
double Ki = 0.1;    // Integralanteil – ggf. anpassen!
double Kd = 0.05;   // Differentialanteil – ggf. anpassen!
double pitchIntegral = 0;
double rollIntegral  = 0;
double lastPitchError = 0;
double lastRollError  = 0;
unsigned long lastPIDTime = 0;

// ***** Automatische Kalibrierung des Gyroskops *****
// Diese Werte werden beim Start (über 100 Samples) ermittelt und von den gemessenen Rohwerten abgezogen.
// Dadurch wird auch kompensiert, wenn das Gyroskop nicht exakt waagrecht angebracht ist.
double gyroPitchOffset = 0.0;
double gyroRollOffset  = 0.0;

// ***** Sensor (MPU6050) *****
Adafruit_MPU6050 mpu;

// ***** Parameter für additive Lenkung *****
// Hier werden feste Korrekturwerte (in RPM/analogem Wert) genutzt.
const int maxRollCorrection = 30;   // Maximale Korrektur für Roll (links/rechts)
const int maxPitchCorrection = 30;  // Maximale Korrektur für Pitch (vorne/hinten)

// ***** Hilfsfunktionen *****

// Map für double-Werte (unverändert)
double mapDouble(double x, double in_min, double in_max, double out_min, double out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

// Clamp-Funktion für Motorwerte (0 bis 255)
int clampMotorSpeed(int speed) {
  if (speed < 0) return 0;
  if (speed > 255) return 255;
  return speed;
}

// Log-Funktion: Gibt Meldungen sowohl über Serial als auch per WebSocket aus
void logMessage(const String &msg) {
  Serial.println(msg);
  ws.textAll(msg);
}

// Stoppt alle Motoren
void stopMotors() {
  ledcWrite(pwmChannel1, 0);
  ledcWrite(pwmChannel2, 0);
  ledcWrite(pwmChannel3, 0);
  ledcWrite(pwmChannel4, 0);
  logMessage("🚨 Motors stopped!");
}

// Automatische Kalibrierung des Gyroskops
// Es werden 100 Samples gesammelt, um Durchschnitts-Offsets für Pitch und Roll zu berechnen.
void calibrateGyro() {
  const int numSamples = 100;
  double sumPitch = 0.0;
  double sumRoll  = 0.0;
  for (int i = 0; i < numSamples; i++) {
    sensors_event_t a, g, temp;
    mpu.getEvent(&a, &g, &temp);
    double rawPitch = atan2(a.acceleration.x, sqrt(a.acceleration.y * a.acceleration.y + a.acceleration.z * a.acceleration.z)) * 57.2958;
    double rawRoll  = atan2(a.acceleration.y, sqrt(a.acceleration.x * a.acceleration.x + a.acceleration.z * a.acceleration.z)) * 57.2958;
    sumPitch += rawPitch;
    sumRoll  += rawRoll;
    delay(10);
  }
  gyroPitchOffset = sumPitch / numSamples;
  gyroRollOffset  = sumRoll / numSamples;
  logMessage("Gyro calibrated: Pitch Offset = " + String(gyroPitchOffset, 2) +
             "°, Roll Offset = " + String(gyroRollOffset, 2) + "°");
}

// ***** Controller-Callbacks *****
void onConnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == nullptr) {
      myControllers[i] = ctl;
      logMessage("🎮 Controller connected!");
      break;
    }
  }
}

void onDisconnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == ctl) {
      myControllers[i] = nullptr;
      logMessage("⚠️ Controller disconnected!");
      stopMotors();
      break;
    }
  }
}

// ***** Controller-Verarbeitung inkl. Lenkung, PID und Kalibrierung *****
void processGamepad(ControllerPtr ctl) {
  lastUpdateTime = millis();

  // Lese Stick-Werte (ca. -511 bis 511)
  int axisX = ctl->axisX();
  int axisY = ctl->axisY();
  logMessage("🎮 Stick: X=" + String(axisX) + ", Y=" + String(axisY));

  // Throttle von 0-1023 auf 0-255 mappen
  int throttleMapped = map(ctl->throttle(), 0, 1023, 0, 255);
  int throttle = throttleMapped;  // Basiswert für die Motoren

  // Hold-Modus: Aktivieren, falls Y-Taste gedrückt
  bool yPressed = (ctl->buttons() & 0x0008);
  int rtPressed = ctl->buttons() & 0x0080;
  if (yPressed && !holdMode) {
    holdMode = true;
    heldMotorSpeed = throttle;
    lastAxisX = axisX;
    lastAxisY = axisY;
    logMessage("🔒 Hold mode activated! Speed locked at " + String(heldMotorSpeed));
  }
  if (holdMode) {
    throttle = heldMotorSpeed;
  }
  if (holdMode && !yPressed) {
    int deltaX = abs(axisX - lastAxisX);
    int deltaY = abs(axisY - lastAxisY);
    if ((deltaX > STICK_DRIFT_THRESHOLD || deltaY > STICK_DRIFT_THRESHOLD) || (ctl->buttons() & ~0x0008)) {
      holdMode = false;
      heldMotorSpeed = 0;
      logMessage("🔓 Hold mode deactivated!");
    }
  }
  if (!holdMode && rtPressed > 0) {
    heldMotorSpeed = throttle;
  }

  // --- Additive Lenkung ---
  // Hier werden feste Korrekturwerte subtrahiert:
  // - Bei links (axisX < 0): Reduktion an Motor 15 & 22 (leftCorrection)
  // - Bei rechts (axisX > 0): Reduktion an Motor 13 & 32 (rightCorrection)
  // - Bei vorne (axisY < 0): Reduktion an Motor 15 & 13 (frontCorrection)
  // - Bei hinten (axisY > 0): Reduktion an Motor 22 & 32 (rearCorrection)
  int leftCorrection = 0, rightCorrection = 0, frontCorrection = 0, rearCorrection = 0;
  if (axisX < 0) {
    leftCorrection = map(abs(axisX), 0, 511, 0, maxRollCorrection);
  } else if (axisX > 0) {
    rightCorrection = map(axisX, 0, 512, 0, maxRollCorrection);
  }
  if (axisY < 0) {
    frontCorrection = map(abs(axisY), 0, 511, 0, maxPitchCorrection);
  } else if (axisY > 0) {
    rearCorrection = map(axisY, 0, 512, 0, maxPitchCorrection);
  }

  // Anforderung:
  // - Bei links: Motor 15 & 22 reduzieren (subtrahiere leftCorrection)
  // - Bei rechts: Motor 13 & 32 reduzieren (subtrahiere rightCorrection)
  // - Bei vorne: Motor 15 & 13 reduzieren (subtrahiere frontCorrection)
  // - Bei hinten: Motor 22 & 32 reduzieren (subtrahiere rearCorrection)
  int motorSpeed15 = clampMotorSpeed(throttle - leftCorrection - frontCorrection); // vorne links
  int motorSpeed22 = clampMotorSpeed(throttle - leftCorrection - rearCorrection);  // hinten links
  int motorSpeed13 = clampMotorSpeed(throttle - rightCorrection - frontCorrection); // vorne rechts
  int motorSpeed32 = clampMotorSpeed(throttle - rightCorrection - rearCorrection);  // hinten rechts

  logMessage("Steering corrections: Left=" + String(leftCorrection) +
             ", Right=" + String(rightCorrection) +
             ", Front=" + String(frontCorrection) +
             ", Rear=" + String(rearCorrection));

  // --- PID-Stabilisierung (Pitch & Roll) mit automatischer Kalibrierung ---
  if (throttle > 0) {
    sensors_event_t a, g, temp;
    mpu.getEvent(&a, &g, &temp);

    // Rohwinkel (in Grad) aus den Beschleunigungsdaten
    double rawPitch = atan2(a.acceleration.x, sqrt(a.acceleration.y * a.acceleration.y + a.acceleration.z * a.acceleration.z)) * 57.2958;
    double rawRoll  = atan2(a.acceleration.y, sqrt(a.acceleration.x * a.acceleration.x + a.acceleration.z * a.acceleration.z)) * 57.2958;
    
    // Korrigiere die gemessenen Winkel mit den automatischen Kalibrierungs-Offsets
    double correctedPitch = rawPitch - gyroPitchOffset;
    double correctedRoll  = rawRoll - gyroRollOffset;
    
    logMessage("Gyro: Raw Pitch=" + String(rawPitch,2) + ", Raw Roll=" + String(rawRoll,2) +
               " | Corrected Pitch=" + String(correctedPitch,2) + ", Corrected Roll=" + String(correctedRoll,2));
    
    // Fehlerberechnung (Sollwert = 0°)
    double errorPitch = 0 - correctedPitch;
    double errorRoll  = 0 - correctedRoll;
    
    unsigned long now = millis();
    double dt = (lastPIDTime == 0) ? 0.01 : (now - lastPIDTime) / 1000.0;
    lastPIDTime = now;
    
    pitchIntegral += errorPitch * dt;
    rollIntegral  += errorRoll * dt;
    
    double pitchDerivative = (errorPitch - lastPitchError) / dt;
    double rollDerivative  = (errorRoll - lastRollError) / dt;
    lastPitchError = errorPitch;
    lastRollError  = errorRoll;
    
    double pidPitch = Kp * errorPitch + Ki * pitchIntegral + Kd * pitchDerivative;
    double pidRoll  = Kp * errorRoll  + Ki * rollIntegral  + Kd * rollDerivative;
    
    // Aufteilung der PID-Korrektur:
    // - Motor 15 (vorne links): +pidPitch + pidRoll
    // - Motor 22 (hinten links): -pidPitch + pidRoll
    // - Motor 13 (vorne rechts): +pidPitch - pidRoll
    // - Motor 32 (hinten rechts): -pidPitch - pidRoll
    int corr15 = (int)(pidPitch + pidRoll);
    int corr22 = (int)(-pidPitch + pidRoll);
    int corr13 = (int)(pidPitch - pidRoll);
    int corr32 = (int)(-pidPitch - pidRoll);
    
    motorSpeed15 = clampMotorSpeed(motorSpeed15 + corr15);
    motorSpeed22 = clampMotorSpeed(motorSpeed22 + corr22);
    motorSpeed13 = clampMotorSpeed(motorSpeed13 + corr13);
    motorSpeed32 = clampMotorSpeed(motorSpeed32 + corr32);
    
    logMessage("PID corrections: Pitch=" + String(pidPitch,2) + ", Roll=" + String(pidRoll,2));
  }
  
  // Setze die PWM-Ausgänge
  ledcWrite(pwmChannel1, motorSpeed15);
  ledcWrite(pwmChannel2, motorSpeed22);
  ledcWrite(pwmChannel3, motorSpeed13);
  ledcWrite(pwmChannel4, motorSpeed32);
  
  logMessage("Motor Speeds: 15=" + String(motorSpeed15) +
             ", 22=" + String(motorSpeed22) +
             ", 13=" + String(motorSpeed13) +
             ", 32=" + String(motorSpeed32));
}

// ***** WebSocket-Event-Handler *****
void onWsEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  if (type == WS_EVT_CONNECT) {
    Serial.printf("✅ WebSocket client #%u connected\n", client->id());
  } else if (type == WS_EVT_DISCONNECT) {
    Serial.printf("❌ WebSocket client #%u disconnected\n", client->id());
  } else if (type == WS_EVT_DATA) {
    AwsFrameInfo *info = (AwsFrameInfo*)arg;
    if (info->opcode == WS_TEXT) {
      String msg = "";
      for (size_t i = 0; i < len; i++) {
        msg += (char)data[i];
      }
      msg.trim();
      if (msg == "calibrate") {
        calibrateGyro();
        logMessage("🔄 Recalibration triggered via WebSocket.");
      }
    }
  }
}

void setup() {
  Serial.begin(921600);
  
  // Starte den ESP32 als Access Point
  logMessage("🌐 Starting Access Point...");
  WiFi.softAP(apSSID, apPassword);
  IPAddress IP = WiFi.softAPIP();
  logMessage("✅ AP started! Connect to \"" + String(apSSID) + "\" (IP: " + IP.toString() + ")");

  // Webserver & WebSocket einrichten
  ws.onEvent(onWsEvent);
  server.addHandler(&ws);
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
    request->send_P(200, "text/html", R"rawliteral(
      <!DOCTYPE html>
      <html>
      <head>
        <meta charset="utf-8">
        <title>ESP32 Log</title>
      </head>
      <body>
        <h1>ESP32 Log</h1>
        <pre id="log"></pre>
        <script>
          var ws = new WebSocket('ws://' + location.host + '/ws');
          ws.onmessage = function(event) {
            var logElem = document.getElementById('log');
            logElem.innerHTML = event.data + "\n" + logElem.innerHTML;
          };
          setInterval(() => ws.send("ping"), 5000);
        </script>
      </body>
      </html>
    )rawliteral");
  });
  server.begin();
  
  // Bluepad32 initialisieren
  BP32.setup(&onConnectedController, &onDisconnectedController);
  BP32.forgetBluetoothKeys();
  
  // I2C initialisieren und MPU6050 einrichten
  Wire.begin(14, 5);  // SDA, SCL
  if (!mpu.begin()) {
    logMessage("❌ Failed to find MPU6050 chip");
    while (1) { delay(10); }
  }
  mpu.setAccelerometerRange(MPU6050_RANGE_8_G);
  mpu.setGyroRange(MPU6050_RANGE_500_DEG);
  mpu.setFilterBandwidth(MPU6050_BAND_21_HZ);
  logMessage("✅ MPU6050 found and initialized!");
  
  // Automatische Kalibrierung des Gyroskops
  calibrateGyro();
  
  // PWM-Kanäle initialisieren und zuordnen
  ledcSetup(pwmChannel1, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel2, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel3, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel4, pwmFrequency, pwmResolution);
  
  ledcAttachPin(MOTOR_PIN1, pwmChannel1);
  ledcAttachPin(MOTOR_PIN2, pwmChannel2);
  ledcAttachPin(MOTOR_PIN3, pwmChannel3);
  ledcAttachPin(MOTOR_PIN4, pwmChannel4);
  
  logMessage("🚀 Setup complete");
}

void loop() {
  if (BP32.update()) {
    for (auto ctl : myControllers) {
      if (ctl && ctl->isConnected()) {
        processGamepad(ctl);
      }
    }
  }
  
  if (millis() - lastUpdateTime > failsafeTimeout) {
    logMessage("⏳ Failsafe activated, stopping motors!");
    stopMotors();
  }
  
  delay(10);
}

/*
Hinweise:
- Der ESP32 startet als Access Point mit SSID "ESP32_Drone" und Passwort "drone1234".
  Verbinde dich per Handy mit diesem WLAN, um über den Browser (IP: 192.168.4.1, falls nicht anders) die Logmeldungen einzusehen.
- Über den WebSocket (unter /ws) können Logmeldungen empfangen werden. Mit dem Befehl "calibrate" wird eine erneute Gyroskop-Kalibrierung ausgelöst.
*/

//ANMERKUNG: Logging mittels WiFi führt zu starken Instabilität!