
#include <Bluepad32.h>
#include <AUnit.h>
using namespace aunit;

// PIN CONNECTIONS
#define MOTOR_PIN1 15 // Motor 1 an Pin 15 vl
#define MOTOR_PIN2 22 // Motor 2 an Pin 22 ul
#define MOTOR_PIN3 13 // Motor 3 an Pin 13 vr
#define MOTOR_PIN4 32 // Motor 4 an Pin 34 ur
 

//    ^ ->  USB - Port
// 15   13
// 32   22


// PWM-Einstellungen
int pwmChannel1 = 0;
int pwmChannel2 = 1;
int pwmChannel3 = 2;
int pwmChannel4 = 3;
int pwmFrequency = 5000;
int pwmResolution = 8;

// Failsafe-Einstellungen
unsigned long lastUpdateTime = 0; 
const unsigned long failsafeTimeout = 20000; 

// Controller- und Hold-Modus-Variablen
ControllerPtr myControllers[BP32_MAX_GAMEPADS];
bool holdMode = false;
int heldMotorSpeed = 0;

const int STICK_DRIFT_THRESHOLD = 50;  // Adjust based on your controller drift level

bool runTests = false;

// Speichern der letzten bekannten neutralen Stickwerte beim Eintritt in den Hold-Modus
int lastAxisX = 0;
int lastAxisY = 0;

// Unit-Tests für MAC-Adressprüfung
struct MockController {
    uint8_t mac[6];
    const uint8_t* getAddress() { return mac; }
};

//Test für valide MAC-Adresse
test(ValidMACAddress) {
    MockController ctl = {{0x00, 0x1A, 0x7D, 0xDA, 0x71, 0x13}};
    assertTrue(isAuthorizedController(ctl));
}

//Test für invalide MAC-Adresse
test(InvalidMACAddress) {
    MockController ctl = {{0x12, 0x34, 0x56, 0x78, 0x9A, 0xBC}};
    assertFalse(isAuthorizedController(ctl));
}

// Funktion zum Testen der Motorsteuerung
void testMotorControl() {
    Serial.println("Starte Motorsteuerungstest...");
    Serial.println("Motoren werden für 2 Sekunden aktiviert...");

    int testSpeed = 150; // Beispiel-PWM-Wert
    ledcWrite(pwmChannel1, testSpeed);
    ledcWrite(pwmChannel2, testSpeed);
    ledcWrite(pwmChannel3, testSpeed);
    ledcWrite(pwmChannel4, testSpeed);

    delay(2000);

    ledcWrite(pwmChannel1, 0);
    ledcWrite(pwmChannel2, 0);
    ledcWrite(pwmChannel3, 0);
    ledcWrite(pwmChannel4, 0);
    Serial.println("Motorsteuerungstest abgeschlossen.");
}

// Funktion zum Testen der Failsafe-Mechanismus
void testFailsafe() {
    Serial.println("Starte Failsafe-Test...");
    Serial.println("Simuliere Signalverlust...");
    
    lastUpdateTime = millis() - (failsafeTimeout + 1000); 
    delay(500); 

    if (millis() - lastUpdateTime > failsafeTimeout) {
        Serial.println("Failsafe aktiviert! Motoren gestoppt.");
        stopMotors();
    } else {
        Serial.println("Failsafe-Test fehlgeschlagen.");
    }
    
    Serial.println("Failsafe-Test abgeschlossen. Zurück zur normalen Steuerung.");
}

// Autorisierte MAC-Adresse
const uint8_t authorizedControllerMAC[6] = {0x00, 0x1A, 0x7D, 0xDA, 0x71, 0x13}; // Beispiel-Adresse

// Funktion zur Überprüfung, ob der Controller autorisiert ist
bool isAuthorizedController(ControllerPtr ctl) {
    const uint8_t* mac = ctl->getAddress();
    Serial.print("Checking MAC Address: ");
    for (int i = 0; i < 6; i++) {
        Serial.printf("%02X:", mac[i]);
        if (mac[i] != authorizedControllerMAC[i]) {
            Serial.println(" -> Unauthorized!");
            return false;
        }
    }
    Serial.println(" -> Authorized!");
    return true;
} 

// Funktion, um die Motoren zu stoppen
void stopMotors() {
  ledcWrite(pwmChannel1, 0);
  ledcWrite(pwmChannel2, 0);
  ledcWrite(pwmChannel3, 0);
  ledcWrite(pwmChannel4, 0);
  Serial.println("Motors stopped");
}

// Funktion, die aufgerufen wird, wenn ein Controller verbunden wird
void onConnectedController(ControllerPtr ctl) {
  if (isAuthorizedController(ctl)) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == nullptr) {
      myControllers[i] = ctl;
      Serial.println("Controller connected");
      break;
    }
  }
}
else {
        Serial.println("Unauthorized controller attempted to connect");
        ctl->disconnect(); //Trennen des nicht autorisierten Controllers
  }
}

// Funktion, die aufgerufen wird, wenn ein Controller getrennt wird
void onDisconnectedController(ControllerPtr ctl) {
  for (int i = 0; i < BP32_MAX_GAMEPADS; i++) {
    if (myControllers[i] == ctl) {
      myControllers[i] = nullptr;
      Serial.println("Controller disconnected");
      break;
    }
  }
  stopMotors(); 
}

// Funktion zur Verarbeitung der Gamepad-Eingaben
void processGamepad(ControllerPtr ctl) {
  lastUpdateTime = millis();

  // Lese die X- und Y-Achse des linken Analogsticks
  int axisX = ctl->axisX();  // -511 bis 511
  int axisY = ctl->axisY();  // -511 bis 511

  // Prüfe, ob die Y-Taste gedrückt wurde
  bool yPressed = (ctl->buttons() & 0x0008);

  // Prüfe, ob die RT-Taste gedrückt wurde
  int rtPressed = ctl->buttons() & 0x0080;
  int motorSpeed = map(ctl->throttle(), 0, 1023, 0, 255);

  // **Hold-Modus aktivieren, wenn Y-Taste gedrückt wird**
  if (yPressed && !holdMode) {
    holdMode = true;
    heldMotorSpeed = motorSpeed;
    lastAxisX = axisX;
    lastAxisY = axisY;
    Serial.println("Hold-Modus aktiviert: Motoren drehen weiter");
  }

  // **Falls Hold-Modus aktiv ist, Motor-Speed konstant halten**
  if (holdMode) {
    motorSpeed = heldMotorSpeed;
  }

  // **Falls neuer Input erkannt wird, Hold-Modus beenden**
  if (holdMode && !yPressed) {
    int deltaX = abs(axisX - lastAxisX);
    int deltaY = abs(axisY - lastAxisY);
    
    bool significantMovement = (deltaX > STICK_DRIFT_THRESHOLD || deltaY > STICK_DRIFT_THRESHOLD);
    
    if (significantMovement || (ctl->buttons() & ~0x0008)) {
      holdMode = false;
      heldMotorSpeed = 0;
      Serial.println("Neuer Input erkannt: Hold-Modus deaktiviert");
    }
  }

  // **Falls RT-Taste wieder gedrückt wird, übernehme RT-Speed**
  if (!holdMode && rtPressed > 0) {
    heldMotorSpeed = motorSpeed;
  }

  // Motoren werden maximal um 30% reduziert
  // Später anpassen
  double maxReduction = 1 - 0.3;

  // axisX (-511 - 512)
  double rightSpeed = (axisX >= 0) ? mapDouble((double)axisX, 0.0, 512.0, 1.0, maxReduction) : 1.0;
  double leftSpeed  = (axisX <  0) ? mapDouble((double)axisX, 0.0, -511.0, 1.0, maxReduction) : 1.0;

  // axisY (-511 - 512)
  double upSpeed   = (axisY <  0) ? mapDouble((double)axisY, 0.0, -511.0, 1.0, maxReduction) : 1.0;
  double downSpeed = (axisY >= 0) ? mapDouble((double)axisY, 0.0, 512.0, 1.0, maxReduction) : 1.0;

  //    ^
  // m1   m3
  // m2   m4

  // Je nach dem wo die Motoren platziert sind, speed multiplikation angepasst werden (130-133)
  int motorSpeed1 = (double)motorSpeed * leftSpeed * upSpeed;
  int motorSpeed2 = (double)motorSpeed * leftSpeed * downSpeed;
  int motorSpeed3 = (double)motorSpeed * rightSpeed * upSpeed;
  int motorSpeed4 = (double)motorSpeed * rightSpeed * downSpeed;

  //Serial.printf("axisX: %d, axisY: %d ", axisX, axisY);
  Serial.printf("Left: %.2f, Up: %.2f, Down: %.2f, Right: %.2f\n", leftSpeed, upSpeed, downSpeed, rightSpeed);

  // Setze die PWM für alle Motoren basierend auf der Bewegung
  ledcWrite(pwmChannel1, motorSpeed1);
  ledcWrite(pwmChannel2, motorSpeed2);
  ledcWrite(pwmChannel3, motorSpeed3);
  ledcWrite(pwmChannel4, motorSpeed4);

  // Debug-Ausgabe für Motorsteuerung
  Serial.print("Motor Speed (Hold-Modus ");
  Serial.print(holdMode ? "aktiv" : "inaktiv");
  Serial.print("): ");
  Serial.println(motorSpeed);
}

// Setup-Funktion
void setup() {
  Serial.begin(921600);
  Serial.println("System ready. Type 't' to run unit tests, 'm' to test motors, 'f' to test failsafe.");
  
  // Initialisiere die PWM-Kanäle
  ledcSetup(pwmChannel1, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel2, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel3, pwmFrequency, pwmResolution);
  ledcSetup(pwmChannel4, pwmFrequency, pwmResolution);

  ledcAttachPin(MOTOR_PIN1, pwmChannel1);
  ledcAttachPin(MOTOR_PIN2, pwmChannel2);
  ledcAttachPin(MOTOR_PIN3, pwmChannel3);
  ledcAttachPin(MOTOR_PIN4, pwmChannel4);

  // Bluepad32 initialisieren
  BP32.setup(&onConnectedController, &onDisconnectedController);
  BP32.forgetBluetoothKeys();

  Serial.println("Setup complete");

   if (runTests) { //wenn runTests = True --> Unittests werden ausgeführt
        Serial.println("Running unit tests...");
        TestRunner::run();
    }
}

// Loop-Funktion
void loop() {
  // Unit Tests werden mit eingabe "t" in seriellem Monitor gestartet
  if (Serial.available()) {
        char command = Serial.read();
        if (command == 't' && !runTests) { // Nur starten, wenn Tests nicht aktiv sind
            Serial.println("Starting unit tests...");
            runTests = true;
            TestRunner::run(); // Führt Tests aus
            Serial.println("Tests abgeschlossen. Zurück zur normalen Steuerung.");
            runTests = false; // Nach Tests zurück zur normalen Steuerung
        }
          if (command == 'm') { 
            testMotorControl(); //Kontrolle der Motoren wird getestet
          }
            if (command == 'f') {
              testFailsafe(); Failsafe wird getestet.
            }
          }
  if (!runTests) {
  if (BP32.update()) {
    for (auto ctl : myControllers) {
      if (ctl && ctl->isConnected()) {
        processGamepad(ctl);
      }
    }
  }

  // Überprüfe den Failsafe-Timeout
  if (millis() - lastUpdateTime > failsafeTimeout) {
    Serial.println("Failsafe triggered, stopping motors...");
    stopMotors();
  }

  delay(10);
  }
}
double mapDouble(double x, double in_min, double in_max, double out_min, double out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

/*
http://drazzy.com/package_drazzy.com_index.json
https://raw.githubusercontent.com/ricardoquesada/esp32-arduino-lib-builder/master/bluepad32_files/package_esp32_bluepad32_index.json
*/